@@ -15,6 +15,7 @@
 #  See the License for the specific language governing permissions and
 # limitations under the License.
 ################################################################################
+import base64
 import collections
 from abc import ABC, abstractmethod
 from enum import Enum
@@ -29,8 +30,9 @@ from typing import List, Tuple, Any, Dict, Collection
 
 from pyflink.datastream import ReduceFunction
 from pyflink.datastream.functions import AggregateFunction
-from pyflink.fn_execution.beam import beam_coder_impl_slow
-from pyflink.fn_execution.coders import TimeWindowCoder, CountWindowCoder
+from pyflink.datastream.state import StateTtlConfig
+from pyflink.fn_execution.beam.beam_coders import FlinkCoder
+from pyflink.fn_execution.coders import FieldCoder
 from pyflink.fn_execution.internal_state import InternalKvState, N, InternalValueState, \
     InternalListState, InternalReducingState, InternalMergingState, InternalAggregatingState, \
     InternalMapState
@@ -85,6 +87,9 @@ class LRUCache(object):
     def __iter__(self):
         return iter(self._cache.values())
 
+    def __contains__(self, key):
+        return key in self._cache
+
 
 class SynchronousKvRuntimeState(InternalKvState, ABC):
     """
@@ -96,6 +101,8 @@ class SynchronousKvRuntimeState(InternalKvState, ABC):
         self._remote_state_backend = remote_state_backend
         self._internal_state = None
         self.namespace = None
+        self._ttl_config = None
+        self._cache_type = SynchronousKvRuntimeState.CacheType.ENABLE_READ_WRITE_CACHE
 
     def set_current_namespace(self, namespace: N) -> None:
         if namespace == self.namespace:
@@ -106,10 +113,26 @@ class SynchronousKvRuntimeState(InternalKvState, ABC):
         self.namespace = namespace
         self._internal_state = None
 
+    def enable_time_to_live(self, ttl_config: StateTtlConfig):
+        self._ttl_config = ttl_config
+        if ttl_config.get_state_visibility() == StateTtlConfig.StateVisibility.NeverReturnExpired:
+            self._cache_type = SynchronousKvRuntimeState.CacheType.DISABLE_CACHE
+        elif ttl_config.get_update_type() == StateTtlConfig.UpdateType.OnReadAndWrite:
+            self._cache_type = SynchronousKvRuntimeState.CacheType.ENABLE_WRITE_CACHE
+
+        if self._cache_type != SynchronousKvRuntimeState.CacheType.ENABLE_READ_WRITE_CACHE:
+            # disable read cache
+            self._remote_state_backend._state_handler._state_cache._cache._max_entries = 0
+
     @abstractmethod
     def get_internal_state(self):
         pass
 
+    class CacheType(Enum):
+        DISABLE_CACHE = 0
+        ENABLE_WRITE_CACHE = 1
+        ENABLE_READ_WRITE_CACHE = 2
+
 
 class SynchronousBagKvRuntimeState(SynchronousKvRuntimeState, ABC):
     """
@@ -122,9 +145,16 @@ class SynchronousBagKvRuntimeState(SynchronousKvRuntimeState, ABC):
     def get_internal_state(self):
         if self._internal_state is None:
             self._internal_state = self._remote_state_backend._get_internal_bag_state(
-                self.name, self.namespace, self._value_coder)
+                self.name, self.namespace, self._value_coder, self._ttl_config)
         return self._internal_state
 
+    def _maybe_clear_write_cache(self):
+        if self._cache_type == SynchronousKvRuntimeState.CacheType.DISABLE_CACHE or \
+                self._remote_state_backend._state_cache_size <= 0:
+            self._internal_state.commit()
+            self._internal_state._cleared = False
+            self._internal_state._added_elements = []
+
 
 class SynchronousValueRuntimeState(SynchronousBagKvRuntimeState, InternalValueState):
     """
@@ -143,6 +173,7 @@ class SynchronousValueRuntimeState(SynchronousBagKvRuntimeState, InternalValueSt
         self.get_internal_state()
         self._internal_state.clear()
         self._internal_state.add(value)
+        self._maybe_clear_write_cache()
 
     def clear(self) -> None:
         self.get_internal_state().clear()
@@ -158,7 +189,7 @@ class SynchronousMergingRuntimeState(SynchronousBagKvRuntimeState, InternalMergi
             name, value_coder, remote_state_backend)
 
     def merge_namespaces(self, target: N, sources: Collection[N]) -> None:
-        self._remote_state_backend.merge_namespaces(self, target, sources)
+        self._remote_state_backend.merge_namespaces(self, target, sources, self._ttl_config)
 
 
 class SynchronousListRuntimeState(SynchronousMergingRuntimeState, InternalListState):
@@ -171,16 +202,19 @@ class SynchronousListRuntimeState(SynchronousMergingRuntimeState, InternalListSt
 
     def add(self, v):
         self.get_internal_state().add(v)
+        self._maybe_clear_write_cache()
 
     def get(self):
         return self.get_internal_state().read()
 
     def add_all(self, values):
         self.get_internal_state()._added_elements.extend(values)
+        self._maybe_clear_write_cache()
 
     def update(self, values):
         self.clear()
         self.add_all(values)
+        self._maybe_clear_write_cache()
 
     def clear(self):
         self.get_internal_state().clear()
@@ -207,6 +241,7 @@ class SynchronousReducingRuntimeState(SynchronousMergingRuntimeState, InternalRe
         else:
             self._internal_state.clear()
             self._internal_state.add(self._reduce_function.reduce(current_value, v))
+        self._maybe_clear_write_cache()
 
     def get(self):
         for i in self.get_internal_state().read():
@@ -241,6 +276,7 @@ class SynchronousAggregatingRuntimeState(SynchronousMergingRuntimeState, Interna
         accumulator = self._agg_function.add(v, accumulator)
         self._internal_state.clear()
         self._internal_state.add(accumulator)
+        self._maybe_clear_write_cache()
 
     def get(self):
         accumulator = self._get_accumulator()
@@ -354,18 +390,18 @@ class CachingMapStateHandler(object):
         else:
             return self._context.bundle_cache_token
 
-    def blocking_get(self, state_key, map_key, map_key_coder, map_value_coder):
+    def blocking_get(self, state_key, map_key, map_key_encoder, map_value_decoder):
         cache_token = self._get_cache_token()
         if not cache_token:
             # cache disabled / no cache token, request from remote directly
-            return self._get_raw(state_key, map_key, map_key_coder, map_value_coder)
+            return self._get_raw(state_key, map_key, map_key_encoder, map_value_decoder)
 
         # lookup cache first
         cache_state_key = self._convert_to_cache_key(state_key)
         cached_map_state = self._state_cache.get(cache_state_key, cache_token)
         if cached_map_state is None:
             # request from remote
-            exists, value = self._get_raw(state_key, map_key, map_key_coder, map_value_coder)
+            exists, value = self._get_raw(state_key, map_key, map_key_encoder, map_value_decoder)
             cached_map_state = CachedMapState(self._max_cached_map_key_entries)
             cached_map_state.put(map_key, (exists, value))
             self._state_cache.put(cache_state_key, cache_token, cached_map_state)
@@ -377,13 +413,15 @@ class CachingMapStateHandler(object):
                     return False, None
 
                 # request from remote
-                exists, value = self._get_raw(state_key, map_key, map_key_coder, map_value_coder)
+                exists, value = self._get_raw(
+                    state_key, map_key, map_key_encoder, map_value_decoder)
                 cached_map_state.put(map_key, (exists, value))
                 return exists, value
             else:
                 return cached_value
 
-    def lazy_iterator(self, state_key, iterate_type, map_key_coder, map_value_coder, iterated_keys):
+    def lazy_iterator(self, state_key, iterate_type, map_key_decoder, map_value_decoder,
+                      iterated_keys):
         cache_token = self._get_cache_token()
         if cache_token:
             # check if the data in the read cache can be used
@@ -396,7 +434,10 @@ class CachingMapStateHandler(object):
         # request from remote
         last_iterator_token = IteratorToken.NOT_START
         current_batch, iterator_token = self._iterate_raw(
-            state_key, iterate_type, last_iterator_token, map_key_coder, map_value_coder)
+            state_key, iterate_type,
+            last_iterator_token,
+            map_key_decoder,
+            map_value_decoder)
 
         if cache_token and \
                 iterator_token == IteratorToken.FINISHED and \
@@ -414,8 +455,8 @@ class CachingMapStateHandler(object):
         return self._lazy_remote_iterator(
             state_key,
             iterate_type,
-            map_key_coder,
-            map_value_coder,
+            map_key_decoder,
+            map_value_decoder,
             iterated_keys,
             iterator_token,
             current_batch)
@@ -424,8 +465,8 @@ class CachingMapStateHandler(object):
             self,
             state_key,
             iterate_type,
-            map_key_coder,
-            map_value_coder,
+            map_key_decoder,
+            map_value_decoder,
             iterated_keys,
             iterator_token,
             current_batch):
@@ -438,7 +479,11 @@ class CachingMapStateHandler(object):
                 if iterator_token == IteratorToken.FINISHED:
                     break
                 current_batch, iterator_token = self._iterate_raw(
-                    state_key, iterate_type, iterator_token, map_key_coder, map_value_coder)
+                    state_key,
+                    iterate_type,
+                    iterator_token,
+                    map_key_decoder,
+                    map_value_decoder)
         elif iterate_type == IterateType.VALUES:
             while True:
                 for key, value in current_batch.items():
@@ -448,7 +493,11 @@ class CachingMapStateHandler(object):
                 if iterator_token == IteratorToken.FINISHED:
                     break
                 current_batch, iterator_token = self._iterate_raw(
-                    state_key, iterate_type, iterator_token, map_key_coder, map_value_coder)
+                    state_key,
+                    iterate_type,
+                    iterator_token,
+                    map_key_decoder,
+                    map_value_decoder)
         elif iterate_type == IterateType.ITEMS:
             while True:
                 for key, value in current_batch.items():
@@ -458,11 +507,16 @@ class CachingMapStateHandler(object):
                 if iterator_token == IteratorToken.FINISHED:
                     break
                 current_batch, iterator_token = self._iterate_raw(
-                    state_key, iterate_type, iterator_token, map_key_coder, map_value_coder)
+                    state_key,
+                    iterate_type,
+                    iterator_token,
+                    map_key_decoder,
+                    map_value_decoder)
         else:
             raise Exception("Unsupported iterate type: %s" % iterate_type)
 
-    def extend(self, state_key, items: List[Tuple[int, Any, Any]], map_key_coder, map_value_coder):
+    def extend(self, state_key, items: List[Tuple[int, Any, Any]],
+               map_key_encoder, map_value_encoder):
         cache_token = self._get_cache_token()
         if cache_token:
             # Cache lookup
@@ -483,8 +537,8 @@ class CachingMapStateHandler(object):
         return self._append_raw(
             state_key,
             items,
-            map_key_coder,
-            map_value_coder)
+            map_key_encoder,
+            map_value_encoder)
 
     def check_empty(self, state_key):
         cache_token = self._get_cache_token()
@@ -534,16 +588,16 @@ class CachingMapStateHandler(object):
         else:
             raise Exception("Unknown response flag: " + str(data[0]))
 
-    def _get_raw(self, state_key, map_key, map_key_coder, map_value_coder):
+    def _get_raw(self, state_key, map_key, map_key_encoder, map_value_decoder):
         output_stream = coder_impl.create_OutputStream()
         output_stream.write_byte(self.GET_FLAG)
-        map_key_coder.encode_to_stream(map_key, output_stream, True)
+        map_key_encoder(map_key, output_stream)
         continuation_token = output_stream.get()
         data, response_token = self._underlying.get_raw(state_key, continuation_token)
         input_stream = coder_impl.create_InputStream(data)
         result_flag = input_stream.read_byte()
         if result_flag == self.EXIST_FLAG:
-            return True, map_value_coder.decode_from_stream(input_stream, True)
+            return True, map_value_decoder(input_stream)
         elif result_flag == self.IS_NONE_FLAG:
             return True, None
         elif result_flag == self.NOT_EXIST_FLAG:
@@ -551,7 +605,8 @@ class CachingMapStateHandler(object):
         else:
             raise Exception("Unknown response flag: " + str(result_flag))
 
-    def _iterate_raw(self, state_key, iterate_type, iterator_token, map_key_coder, map_value_coder):
+    def _iterate_raw(self, state_key, iterate_type, iterator_token,
+                     map_key_decoder, map_value_decoder):
         output_stream = coder_impl.create_OutputStream()
         output_stream.write_byte(self.ITERATE_FLAG)
         output_stream.write_byte(iterate_type.value)
@@ -583,10 +638,10 @@ class CachingMapStateHandler(object):
             # decode both key and value
             current_batch = {}
             while input_stream.size() > 0:
-                key = map_key_coder.decode_from_stream(input_stream, True)
+                key = map_key_decoder(input_stream)
                 is_not_none = input_stream.read_byte()
                 if is_not_none:
-                    value = map_value_coder.decode_from_stream(input_stream, True)
+                    value = map_value_decoder(input_stream)
                 else:
                     value = None
                 current_batch[key] = value
@@ -594,11 +649,11 @@ class CachingMapStateHandler(object):
             # only decode key
             current_batch = []
             while input_stream.size() > 0:
-                key = map_key_coder.decode_from_stream(input_stream, True)
+                key = map_key_decoder(input_stream)
                 current_batch.append(key)
         return current_batch, new_iterator_token
 
-    def _append_raw(self, state_key, items, map_key_coder, map_value_coder):
+    def _append_raw(self, state_key, items, map_key_encoder, map_value_encoder):
         output_stream = coder_impl.create_OutputStream()
         output_stream.write_bigendian_int32(len(items))
         for request_flag, map_key, map_value in items:
@@ -606,13 +661,13 @@ class CachingMapStateHandler(object):
             # Not all the coder impls will serialize the length of bytes when we set the "nested"
             # param to "True", so we need to encode the length of bytes manually.
             tmp_out = coder_impl.create_OutputStream()
-            map_key_coder.encode_to_stream(map_key, tmp_out, True)
+            map_key_encoder(map_key, tmp_out)
             serialized_data = tmp_out.get()
             output_stream.write_bigendian_int32(len(serialized_data))
             output_stream.write(serialized_data)
             if request_flag == self.SET_VALUE:
                 tmp_out = coder_impl.create_OutputStream()
-                map_value_coder.encode_to_stream(map_value, tmp_out, True)
+                map_value_encoder(map_value, tmp_out)
                 serialized_data = tmp_out.get()
                 output_stream.write_bigendian_int32(len(serialized_data))
                 output_stream.write(serialized_data)
@@ -623,7 +678,7 @@ class CachingMapStateHandler(object):
         return state_key.SerializeToString()
 
 
-class RemovableConcatIterator(collections.Iterator):
+class RemovableConcatIterator(collections.abc.Iterator):
 
     def __init__(self, internal_map_state, first, second):
         self._first = first
@@ -684,20 +739,19 @@ class InternalSynchronousMapRuntimeState(object):
         self._map_state_handler = map_state_handler
         self._state_key = state_key
         self._map_key_coder = map_key_coder
-        # TODO: Replace slow WindowCoderImpl with fast WindowCoderImpl
-        if isinstance(map_key_coder, TimeWindowCoder):
-            self._map_key_coder_impl = beam_coder_impl_slow.TimeWindowCoderImpl()
-        elif isinstance(map_key_coder, CountWindowCoder):
-            self._map_key_coder_impl = beam_coder_impl_slow.CountWindowCoderImpl()
+        if isinstance(map_key_coder, FieldCoder):
+            map_key_coder_impl = FlinkCoder(map_key_coder).get_impl()
         else:
-            self._map_key_coder_impl = map_key_coder.get_impl()
+            map_key_coder_impl = map_key_coder.get_impl()
+        self._map_key_encoder, self._map_key_decoder = \
+            self._get_encoder_and_decoder(map_key_coder_impl)
         self._map_value_coder = map_value_coder
-        if isinstance(map_value_coder, TimeWindowCoder):
-            self._map_value_coder_impl = beam_coder_impl_slow.TimeWindowCoderImpl()
-        elif isinstance(map_key_coder, CountWindowCoder):
-            self._map_value_coder_impl = beam_coder_impl_slow.CountWindowCoderImpl()
+        if isinstance(map_value_coder, FieldCoder):
+            map_value_coder_impl = FlinkCoder(map_value_coder).get_impl()
         else:
-            self._map_value_coder_impl = map_value_coder.get_impl()
+            map_value_coder_impl = map_value_coder.get_impl()
+        self._map_value_encoder, self._map_value_decoder = \
+            self._get_encoder_and_decoder(map_value_coder_impl)
         self._write_cache = dict()
         self._max_write_cache_entries = max_write_cache_entries
         self._is_empty = None
@@ -716,7 +770,7 @@ class InternalSynchronousMapRuntimeState(object):
         if self._cleared:
             return None
         exists, value = self._map_state_handler.blocking_get(
-            self._state_key, map_key, self._map_key_coder_impl, self._map_value_coder_impl)
+            self._state_key, map_key, self._map_key_encoder, self._map_value_decoder)
         if exists:
             return value
         else:
@@ -807,7 +861,7 @@ class InternalSynchronousMapRuntimeState(object):
                     append_items.append((CachingMapStateHandler.DELETE, map_key, None))
             self._write_cache.clear()
             to_await = self._map_state_handler.extend(
-                self._state_key, append_items, self._map_key_coder_impl, self._map_value_coder_impl)
+                self._state_key, append_items, self._map_key_encoder, self._map_value_encoder)
         if to_await:
             to_await.get()
         self._write_cache.clear()
@@ -824,10 +878,16 @@ class InternalSynchronousMapRuntimeState(object):
             return self._map_state_handler.lazy_iterator(
                 self._state_key,
                 iterate_type,
-                self._map_key_coder_impl,
-                self._map_value_coder_impl,
+                self._map_key_decoder,
+                self._map_value_decoder,
                 self._write_cache)
 
+    @staticmethod
+    def _get_encoder_and_decoder(coder):
+        encoder = partial(coder.encode_to_stream, nested=True)
+        decoder = partial(coder.decode_from_stream, nested=True)
+        return encoder, decoder
+
 
 class SynchronousMapRuntimeState(SynchronousKvRuntimeState, InternalMapState):
 
@@ -843,7 +903,12 @@ class SynchronousMapRuntimeState(SynchronousKvRuntimeState, InternalMapState):
     def get_internal_state(self):
         if self._internal_state is None:
             self._internal_state = self._remote_state_backend._get_internal_map_state(
-                self.name, self.namespace, self._map_key_coder, self._map_value_coder)
+                self.name,
+                self.namespace,
+                self._map_key_coder,
+                self._map_value_coder,
+                self._ttl_config,
+                self._cache_type)
         return self._internal_state
 
     def get(self, key):
@@ -894,8 +959,7 @@ class RemoteKeyedStateBackend(object):
         self._state_handler = state_handler
         self._map_state_handler = CachingMapStateHandler(
             state_handler, map_state_read_cache_size)
-        from pyflink.fn_execution.coders import FlattenRowCoder
-        self._key_coder_impl = FlattenRowCoder(key_coder._field_coders).get_impl()
+        self._key_coder_impl = key_coder.get_impl()
         self.namespace_coder = namespace_coder
         if namespace_coder:
             self._namespace_coder_impl = namespace_coder.get_impl()
@@ -915,31 +979,47 @@ class RemoteKeyedStateBackend(object):
                 side_input_id="clear_iterators",
                 key=self._encoded_current_key))
 
-    def get_list_state(self, name, element_coder):
+    def get_list_state(self, name, element_coder, ttl_config=None):
         return self._wrap_internal_bag_state(
-            name, element_coder, SynchronousListRuntimeState, SynchronousListRuntimeState)
+            name,
+            element_coder,
+            SynchronousListRuntimeState,
+            SynchronousListRuntimeState,
+            ttl_config)
 
-    def get_value_state(self, name, value_coder):
+    def get_value_state(self, name, value_coder, ttl_config=None):
         return self._wrap_internal_bag_state(
-            name, value_coder, SynchronousValueRuntimeState, SynchronousValueRuntimeState)
+            name,
+            value_coder,
+            SynchronousValueRuntimeState,
+            SynchronousValueRuntimeState,
+            ttl_config)
 
-    def get_map_state(self, name, map_key_coder, map_value_coder):
+    def get_map_state(self, name, map_key_coder, map_value_coder, ttl_config=None):
         if name in self._all_states:
             self.validate_map_state(name, map_key_coder, map_value_coder)
             return self._all_states[name]
         map_state = SynchronousMapRuntimeState(name, map_key_coder, map_value_coder, self)
+        if ttl_config is not None:
+            map_state.enable_time_to_live(ttl_config)
         self._all_states[name] = map_state
         return map_state
 
-    def get_reducing_state(self, name, coder, reduce_function):
+    def get_reducing_state(self, name, coder, reduce_function, ttl_config=None):
         return self._wrap_internal_bag_state(
-            name, coder, SynchronousReducingRuntimeState,
-            partial(SynchronousReducingRuntimeState, reduce_function=reduce_function))
+            name,
+            coder,
+            SynchronousReducingRuntimeState,
+            partial(SynchronousReducingRuntimeState, reduce_function=reduce_function),
+            ttl_config)
 
-    def get_aggregating_state(self, name, coder, agg_function):
+    def get_aggregating_state(self, name, coder, agg_function, ttl_config=None):
         return self._wrap_internal_bag_state(
-            name, coder, SynchronousAggregatingRuntimeState,
-            partial(SynchronousAggregatingRuntimeState, agg_function=agg_function))
+            name,
+            coder,
+            SynchronousAggregatingRuntimeState,
+            partial(SynchronousAggregatingRuntimeState, agg_function=agg_function),
+            ttl_config)
 
     def validate_state(self, name, coder, expected_type):
         if name in self._all_states:
@@ -960,15 +1040,18 @@ class RemoteKeyedStateBackend(object):
                     state._map_value_coder != map_value_coder:
                 raise Exception("State name corrupted: %s" % name)
 
-    def _wrap_internal_bag_state(self, name, element_coder, wrapper_type, wrap_method):
+    def _wrap_internal_bag_state(
+            self, name, element_coder, wrapper_type, wrap_method, ttl_config):
         if name in self._all_states:
             self.validate_state(name, element_coder, wrapper_type)
             return self._all_states[name]
         wrapped_state = wrap_method(name, element_coder, self)
+        if ttl_config is not None:
+            wrapped_state.enable_time_to_live(ttl_config)
         self._all_states[name] = wrapped_state
         return wrapped_state
 
-    def _get_internal_bag_state(self, name, namespace, element_coder):
+    def _get_internal_bag_state(self, name, namespace, element_coder, ttl_config):
         encoded_namespace = self._encode_namespace(namespace)
         cached_state = self._internal_state_cache.get(
             (name, self._encoded_current_key, encoded_namespace))
@@ -978,51 +1061,64 @@ class RemoteKeyedStateBackend(object):
         # at once. The internal state cache is only updated when the current key changes.
         # The reason is that the state cache size may be smaller that the count of activated
         # state (i.e. the state with current key).
+        if isinstance(element_coder, FieldCoder):
+            element_coder = FlinkCoder(element_coder)
         state_spec = userstate.BagStateSpec(name, element_coder)
-        internal_state = self._create_bag_state(state_spec, encoded_namespace)
+        internal_state = self._create_bag_state(state_spec, encoded_namespace, ttl_config)
         return internal_state
 
-    def _get_internal_map_state(self, name, namespace, map_key_coder, map_value_coder):
+    def _get_internal_map_state(
+            self, name, namespace, map_key_coder, map_value_coder, ttl_config, cache_type):
         encoded_namespace = self._encode_namespace(namespace)
         cached_state = self._internal_state_cache.get(
             (name, self._encoded_current_key, encoded_namespace))
         if cached_state is not None:
             return cached_state
         internal_map_state = self._create_internal_map_state(
-            name, encoded_namespace, map_key_coder, map_value_coder)
+            name, encoded_namespace, map_key_coder, map_value_coder, ttl_config, cache_type)
         return internal_map_state
 
-    def _create_bag_state(self, state_spec: userstate.StateSpec, encoded_namespace) \
+    def _create_bag_state(self, state_spec: userstate.StateSpec, encoded_namespace, ttl_config) \
             -> userstate.AccumulatingRuntimeState:
         if isinstance(state_spec, userstate.BagStateSpec):
             bag_state = SynchronousBagRuntimeState(
                 self._state_handler,
                 state_key=self.get_bag_state_key(
-                    state_spec.name, self._encoded_current_key, encoded_namespace),
+                    state_spec.name, self._encoded_current_key, encoded_namespace, ttl_config),
                 value_coder=state_spec.coder)
             return bag_state
         else:
             raise NotImplementedError(state_spec)
 
-    def _create_internal_map_state(self, name, encoded_namespace, map_key_coder, map_value_coder):
+    def _create_internal_map_state(
+            self, name, encoded_namespace, map_key_coder, map_value_coder, ttl_config, cache_type):
         # Currently the `beam_fn_api.proto` does not support MapState, so we use the
         # the `MultimapSideInput` message to mark the state as a MapState for now.
+        from pyflink.fn_execution.flink_fn_execution_pb2 import StateDescriptor
+        state_proto = StateDescriptor()
+        state_proto.state_name = name
+        if ttl_config is not None:
+            state_proto.state_ttl_config.CopyFrom(ttl_config._to_proto())
         state_key = beam_fn_api_pb2.StateKey(
             multimap_side_input=beam_fn_api_pb2.StateKey.MultimapSideInput(
                 transform_id="",
                 window=encoded_namespace,
-                side_input_id=name,
+                side_input_id=base64.b64encode(state_proto.SerializeToString()),
                 key=self._encoded_current_key))
+        if cache_type == SynchronousKvRuntimeState.CacheType.DISABLE_CACHE:
+            write_cache_size = 0
+        else:
+            write_cache_size = self._map_state_write_cache_size
         return InternalSynchronousMapRuntimeState(
             self._map_state_handler,
             state_key,
             map_key_coder,
             map_value_coder,
-            self._map_state_write_cache_size)
+            write_cache_size)
 
     def _encode_namespace(self, namespace):
         if namespace is not None:
-            encoded_namespace = self._namespace_coder_impl.encode_nested(namespace)
+            encoded_namespace = self._namespace_coder_impl.encode(namespace)
         else:
             encoded_namespace = b''
         return encoded_namespace
@@ -1044,7 +1140,7 @@ class RemoteKeyedStateBackend(object):
             state_obj.namespace = None
             state_obj._internal_state = None
         self._current_key = key
-        self._encoded_current_key = self._key_coder_impl.encode_nested(self._current_key)
+        self._encoded_current_key = self._key_coder_impl.encode(self._current_key)
 
     def get_current_key(self):
         return self._current_key
@@ -1062,17 +1158,18 @@ class RemoteKeyedStateBackend(object):
             self._clear_iterator_mark.multimap_side_input.key = self._encoded_current_key
             self._map_state_handler.clear(self._clear_iterator_mark)
 
-    def merge_namespaces(self, state: SynchronousMergingRuntimeState, target, sources):
+    def merge_namespaces(self, state: SynchronousMergingRuntimeState, target, sources, ttl_config):
+        for source in sources:
+            state.set_current_namespace(source)
+            self.commit_internal_state(state.get_internal_state())
         state.set_current_namespace(target)
         self.commit_internal_state(state.get_internal_state())
         encoded_target_namespace = self._encode_namespace(target)
-        encoded_namespaces = [encoded_target_namespace]
-        for source in sources:
-            encoded_namespaces.append(self._encode_namespace(source))
-        self.clear_state_cache(state, encoded_namespaces)
+        encoded_namespaces = [self._encode_namespace(source) for source in sources]
+        self.clear_state_cache(state, [encoded_target_namespace] + encoded_namespaces)
 
         state_key = self.get_bag_state_key(
-            state.name, self._encoded_current_key, encoded_target_namespace)
+            state.name, self._encoded_current_key, encoded_target_namespace, ttl_config)
         state_key.bag_user_state.transform_id = self.MERGE_NAMESAPCES_MARK
 
         encoded_namespaces_writer = BytesIO()
@@ -1093,18 +1190,22 @@ class RemoteKeyedStateBackend(object):
                     (name, self._encoded_current_key, encoded_namespace))
                 # currently all the SynchronousMergingRuntimeState is based on bag state
                 state_key = self.get_bag_state_key(
-                    name, self._encoded_current_key, encoded_namespace)
+                    name, self._encoded_current_key, encoded_namespace, None)
                 # clear the read cache, the read cache is shared between map state handler and bag
                 # state handler. So we can use the map state handler instead.
                 self._map_state_handler.clear_read_cache(state_key)
 
-    @staticmethod
-    def get_bag_state_key(name, encoded_key, encoded_namespace):
+    def get_bag_state_key(self, name, encoded_key, encoded_namespace, ttl_config):
+        from pyflink.fn_execution.flink_fn_execution_pb2 import StateDescriptor
+        state_proto = StateDescriptor()
+        state_proto.state_name = name
+        if ttl_config is not None:
+            state_proto.state_ttl_config.CopyFrom(ttl_config._to_proto())
         return beam_fn_api_pb2.StateKey(
             bag_user_state=beam_fn_api_pb2.StateKey.BagUserState(
                 transform_id="",
                 window=encoded_namespace,
-                user_state_id=name,
+                user_state_id=base64.b64encode(state_proto.SerializeToString()),
                 key=encoded_key))
 
     @staticmethod
